import os
import numpy as np
import random
import datetime
import torch
__all__ = ['save_state_dict', 'load_state_dict', 'load_state_dict_test']


def save_state_dict(model, optimizer, lr_scheduler, train_loss, val_loss, checkpoint_dir, model_architecture, epoch_idx):
    """Save the state dict of the current training to disk

    Parameters
    ----------
    model : torch model
        trained model to save
    optimizer : torch.optim object
    lr_scheduler: torch.optim.lr_scheduler object
    checkpoint_dir : str or os.path object
        directory into which to save the model
    model_architecture : str
        type of architecture
    epoch : int
        epoch index

    Returns
    -------
    str or os.path object
        path to the saved model

    """
    state = dict(
                 model=model.state_dict(),
                 optimizer=optimizer.state_dict(),
                 lr_scheduler=lr_scheduler.state_dict(),
                 epoch=epoch_idx,
                 train_loss=train_loss,
                 val_loss=val_loss,
                 )
    time_stamp = datetime.datetime.now().strftime("epoch={:d}_%m-%d-%Y_%H:%M".format(epoch_idx))
    model_fname = '{:s}_{:s}.mdl'.format(model_architecture, time_stamp)
    model_path = os.path.join(checkpoint_dir, model_fname)
    torch.save(state, model_path)
    return model_path

def load_state_dict(checkpoint_path, model, optimizer, n_epochs, device, lr_scheduler=None):
    """Load the state dict of the past training

    Parameters
    ----------
    checkpoint_path : str or os.path object
        path of the state dict to load
    model : torch model
        trained model to save
    optimizer : torch.optim object
    lr_scheduler: torch.optim.lr_scheduler object
    n_epochs : int
        total number of epochs to train
    device : torch.device object
        device on which to load the model

    Returns
    -------
    str or os.path object
        path to the saved model

    """
    state = torch.load(checkpoint_path)
    model.load_state_dict(state['model'])
    model.to(device)
    optimizer.load_state_dict(state['optimizer'])
    if lr_scheduler is not None:
        lr_scheduler.load_state_dict(state['lr_scheduler'])
    epoch = state['epoch']
    train_loss = state['train_loss']
    val_loss = state['val_loss']
    print("Loaded weights at {:s}".format(checkpoint_path))
    print("Epoch [{}/{}]: TRAIN Loss: {:.4f}".format(epoch+1, n_epochs, train_loss))
    print("Epoch [{}/{}]: VALID Loss: {:.4f}".format(epoch+1, n_epochs, val_loss))
    return epoch, model, optimizer, train_loss, val_loss

def load_state_dict_test(checkpoint_path, model, n_epochs, device):
    """Load the state dict of the past training

    Parameters
    ----------
    checkpoint_path : str or os.path object
        path of the state dict to load
    model : torch model
        trained model to save
    optimizer : torch.optim object
    lr_scheduler: torch.optim.lr_scheduler object
    n_epochs : int
        total number of epochs to train
    device : torch.device object
        device on which to load the model

    Returns
    -------
    str or os.path object
        path to the saved model

    """
    state = torch.load(checkpoint_path)
    model.load_state_dict(state['model'])
    model.to(device)
    epoch = state['epoch']
    train_loss = state['train_loss']
    val_loss = state['val_loss']
    print("Loaded weights at {:s}".format(checkpoint_path))
    print("Epoch [{}/{}]: TRAIN Loss: {:.4f}".format(epoch+1, n_epochs, train_loss))
    print("Epoch [{}/{}]: VALID Loss: {:.4f}".format(epoch+1, n_epochs, val_loss))
    return model, epoch


import os, sys
import importlib
import warnings
import json
import glob
import numpy as np
import pandas as pd
from addict import Dict
from baobab.sim_utils import add_g1g2_columns

class TrainValConfig:
    """Nested dictionary representing the configuration for H0rton training, h0_inference, visualization, and analysis

    """
    def __init__(self, user_cfg):
        """
        Parameters
        ----------
        user_cfg : dict or Dict
            user-defined configuration
        
        """
        self.__dict__ = Dict(user_cfg)
        self.validate_user_definition()
        self.preset_default()
        self.set_monitoring_cfg()        

    @classmethod
    def from_file(cls, user_cfg_path):
        """Alternative constructor that accepts the path to the user-defined configuration python file

        Parameters
        ----------
        user_cfg_path : str or os.path object
            path to the user-defined configuration python file

        """
        dirname, filename = os.path.split(os.path.abspath(user_cfg_path))
        module_name, ext = os.path.splitext(filename)
        sys.path.append(dirname)
        if ext == '.py':
            #user_cfg_file = map(__import__, module_name)
            #user_cfg = getattr(user_cfg_file, 'cfg')
            user_cfg_script = importlib.import_module(module_name)
            user_cfg = getattr(user_cfg_script, 'cfg')
            return cls(user_cfg)
        elif ext == '.json':
            with open(user_cfg_path, 'r') as f:
                user_cfg_str = f.read()
            user_cfg = Dict(json.loads(user_cfg_str))
            return cls(user_cfg)
        else:
            raise NotImplementedError("This extension is not supported.")

    def validate_user_definition(self):
        """Check to see if the user-defined config is valid

        """
        import h0rton.losses
        if not hasattr(h0rton.losses, self.model.likelihood_class):
            raise TypeError("Likelihood class supplied in cfg doesn't exist.")

    def preset_default(self):
        """Preset default config values

        """
        if 'train_baobab_cfg_path' not in self.data:
            raise ValueError("Must provide training data directory.")
        if 'val_baobab_cfg_path' not in self.data:
            raise ValueError("Must provide validation data directory.")
        # FIXME: doesn't check for contents of baobab config file, just the file names
        if self.data.train_baobab_cfg_path == self.data.val_baobab_cfg_path:
            warnings.warn("You're training and validating on the same dataset.", UserWarning, stacklevel=2)
        if 'float_type' not in self.data:
            self.data.float_type = 'FloatTensor'
            warnings.warn("Float type not provided. Defaulting to float32...")

    def set_monitoring_cfg(self):
        """Set general metadata relevant to network architecture and optimization

        """
        # Data to plot during monitoring
        if self.monitoring.n_plotting > 100:
            warnings.warn("Only plotting allowed max of 100 datapoints during training")
            self.monitoring.n_plotting = 100
        if self.monitoring.n_plotting > self.optim.batch_size:
            raise ValueError("monitoring.n_plotting must be smaller than optim.batch_size")


import numpy as np
import torch
__all__ = ['rescale_01', 'whiten_Y_cols', 'plus_1_log', 'asinh', 'whiten_pixels', 'log_parameterize_Y_cols']

def whiten_pixels(pixels):
    return (pixels - torch.mean(pixels))/torch.std(pixels)

def asinh(x):
    return torch.log(x+(x**2+1)**0.5)

def plus_1_log(linear):
    """Add 1 and take the log10 of an image

    Parameters
    ----------
    linear : torch.Tensor of shape `[X_dim, X_dim]`

    Returns
    -------
    torch.Tensor
        the image of the same input shape, with values now logged

    """
    return torch.log1p(linear)

def rescale_01(unscaled):
    """Rescale an image of unknown range to values between 0 and 1

    Parameters
    ----------
    unscaled : torch.Tensor of shape `[X_dim, X_dim]`

    Returns
    -------
    torch.Tensor
        the image of the same input shape, with values now scaled between 0 and 1

    """
    return (unscaled - unscaled.min())/(unscaled.max() - unscaled.min())

def whiten_Y_cols(df, mean, std, col_names):
    """Whiten (in place) select columns in the given dataframe, i.e. shift and scale then so that they have the desired mean and std

    Parameters
    ----------
    df : pd.DataFrame
    mean : array-like
        target mean
    std : array-like
        target std
    col_names : list
        names of columns to whiten

    """
    df.loc[:, col_names] = (df.loc[:, col_names].values - mean)/std
    #return df

def log_parameterize_Y_cols(df, col_names):
    """Whiten (in place) select columns in the given dataframe, i.e. shift and scale then so that they have the desired mean and std

    Parameters
    ----------
    df : pd.DataFrame
    mean : array-like
        target mean
    std : array-like
        target std
    col_names : list
        names of columns to whiten

    """
    df.loc[:, col_names] = np.log(df.loc[:, col_names].values)


import numpy as np
import mpmath as mp
from astropy.cosmology import FlatLambdaCDM
import baobab.sim_utils.kinematics_utils as kinematics_utils
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.Analysis.td_cosmography import TDCosmography
from scipy.stats import norm
from h0rton.h0_inference import h0_utils

__all__ = ['H0Posterior']

class H0Posterior:
    """Represents the posterior over H0

    """
    required_params = ["lens_mass_center_x", "src_light_center_x","lens_mass_center_y", "src_light_center_y", "lens_mass_gamma", "lens_mass_theta_E", "lens_mass_e1", "lens_mass_e2", "external_shear_gamma1", "external_shear_gamma2", "src_light_R_sersic"]
    def __init__(self, H0_prior, kappa_ext_prior,  kwargs_model, baobab_time_delays, Om0, define_src_pos_wrt_lens, exclude_vel_disp=True, aniso_param_prior=None, kinematics=None, kappa_transformed=True, kwargs_lens_eqn_solver={}):
        """

        Parameters
        ----------
        H0_prior : scipy rv_continuous object
        kappa_ext_prior : scipy rv_continuous object
            prior over the external convergence
        aniso_param_prior : scipy rv_continuous object
            prior over the anisotropy radius, r_ani
        kwargs_model : dict
            dictionary defining which models (parameterizations) are used to define `lens_mass_dict`, `ext_shear_dict`, `ps_dict`
        abcd_ordering_i : list
            ABCD in an increasing dec order if the keys ABCD mapped to values 0123, respectively, e.g. [3, 1, 0, 2] if D (value 3) is lowest, B (value 1) is second lowest
        exclude_vel_disp : bool
            whether to exclude the velocity dispersion likelihood. Default: False
        
        """
        self.H0_prior = H0_prior
        self.kappa_transformed = kappa_transformed
        self.kappa_ext_prior = kappa_ext_prior
        if self.kappa_transformed:
            self.sample_kappa_ext = self.sample_kappa_ext_transformed
        else:
            self.sample_kappa_ext = self.sample_kappa_ext_original
        self.aniso_param_prior = aniso_param_prior
        self.exclude_vel_disp = exclude_vel_disp
        self.kwargs_model = kwargs_model
        self.baobab_time_delays = baobab_time_delays
        self.define_src_pos_wrt_lens = define_src_pos_wrt_lens
        self.kinematics = kinematics
        self.Om0 = Om0 # Omega matter
        self.kwargs_lens_eqn_solver = kwargs_lens_eqn_solver
        self.kwargs_model.update(dict(point_source_model_list=['SOURCE_POSITION']))
        if not self.exclude_vel_disp:
            if self.kinematics is None:
                raise ValueError("kinematics is required to calculate velocity dispersion.")
            if self.kinematics.anisotropy_model == 'analytic':
                self.get_velocity_dispersion = getattr(kinematics_utils, 'velocity_dispersion_analytic')
            else:
                # TODO: currently not available, as BNN does not predict lens light profile
                self.get_velocity_disperison = getattr(kinematics_utils, 'velocity_dispersion_numerical')

    @classmethod
    def from_dict(cls, lens_dict):
        """Initialize H0Posterior from a dictionary

        Parameters
        ----------
        lens_dict : dict
            contains properties required to initialize H0Posterior. See `__init__` method above for the required parameters and their formats.

        """
        return cls(lens_dict.items())

    def set_cosmology_observables(self, z_lens, z_src, measured_td_wrt0, measured_td_err, abcd_ordering_i, true_img_dec, true_img_ra, kappa_ext, measured_vd=None, measured_vd_err=None):
        """Set the cosmology observables for a given lens system, persistent across all the samples for that system

        Parameters
        ----------
        z_lens : float
        z_src : float
        lens_mass_dict : dict
            dict of lens mass kwargs
        ext_shear_dict : dict
            dict of external shear kwargs
        ps_dict : dict
            dict of point source kwargs
        measured_vd : float
            measured velocity dispersion
        measured_vd_err : float
            measurement error of velocity dispersion
        lens_light_R_sersic : float
            effective radius of lens light in arcsec
        measured_td : np.array of shape `[n_images,]`
            the measured time delays in days 
        measured_td_err : float
            the time delay measurement error in days
        abcd_ordering_i : np.array of shape `[n_images,]`
            the image ordering followed by `measured_td` in increasing dec. Example: if the `measured_td` are [a, b, c, d] and the corresponding image dec are [0.3, -0.1, 0.8, 0.4], then `abcd_ordering_i` are [1, 0, 3, 2].
        true_img_dec : np.array of shape `[n_images, ]`
            dec of the true image positions in arcsec
        true_img_ra : np.array of shape `[n_images, ]`
            ra of the true image positions in arcsec

        """
        self.z_lens = z_lens
        self.z_src = z_src
        self.measured_vd = measured_vd
        self.measured_vd_err = measured_vd_err
        self.measured_td_wrt0 = np.array(measured_td_wrt0)
        self.measured_td_err = np.array(measured_td_err)
        self.abcd_ordering_i = abcd_ordering_i
        self.true_img_dec = true_img_dec
        self.true_img_ra = true_img_ra
        self.kappa_ext = kappa_ext
        #self._reorder_measured_td_to_tdlmc()
        # Number of AGN images
        self.n_img = len(measured_td_wrt0) + 1

    def _reorder_measured_td_to_tdlmc(self):
        """Reorder the measured time delays (same for all lens model samples)
        
        Note
        ----
        Unused!

        """
        #print(self.measured_td, self.true_img_dec, self.abcd_ordering_i)
        reordered_measured_td = h0_utils.reorder_to_tdlmc(self.measured_td, np.argsort(self.true_img_dec), self.abcd_ordering_i)
        # Measured time in days (offset from the image with the smallest dec)
        self.measured_td_wrt0 = reordered_measured_td[1:] - reordered_measured_td[0]

    def format_lens_model(self, sample):
        """Set the lens model parameters for a given lens mass model

        Parameters
        ----------
        sample : dict
            a sampled set of lens model parameters

        """
        # Lens mass
        # FIXME: hardcoded for SPEMD
        kwargs_spemd = {'theta_E': sample['lens_mass_theta_E'],
                        'center_x': sample['lens_mass_center_x'], 
                        'center_y': sample['lens_mass_center_y'],
                        'e1': sample['lens_mass_e1'], 
                        'e2': sample['lens_mass_e2'], 
                        'gamma': sample['lens_mass_gamma'],}
        # External shear
        kwargs_shear = {'gamma1': sample['external_shear_gamma1'],
                        'gamma2': sample['external_shear_gamma2'],
                        'ra_0': sample['lens_mass_center_x'],
                        'dec_0': sample['lens_mass_center_y']}
        # AGN point source
        if self.define_src_pos_wrt_lens:
            kwargs_ps = {
            'ra_source': sample['src_light_center_x'] + sample['lens_mass_center_x'],
            'dec_source': sample['src_light_center_y'] + sample['lens_mass_center_y'],
            }
        else:
            kwargs_ps = {
            'ra_source': sample['src_light_center_x'],
            'dec_source': sample['src_light_center_y'],
            }
        kwargs_lens = [kwargs_spemd, kwargs_shear]
        # Raytrace to get point source kwargs in image plane
        #kwargs_img, requires_reordering = self.get_img_pos(kwargs_ps, kwargs_lens)
        # Pre-store for reordering image arrays
        #dec_image = kwargs_img[0]['dec_image']
        #increasing_dec_i = np.argsort(dec_image)

        formatted_lens_model = dict(
                          # TODO: key checking depending on kwargs_model
                          kwargs_lens=kwargs_lens,
                          kwargs_ps=[kwargs_ps],
                          #kwargs_img=kwargs_img,
                          requires_reordering=True, #FIXME: get edge cases
                          #increasing_dec_i=increasing_dec_i,
                          #lens_light_R_sersic=sample['lens_light_R_sersic'],
                          )
        return formatted_lens_model

    def get_img_pos(self, ps_dict, kwargs_lens):
        """Sets the kwargs_ps class attribute as those coresponding to the point source model `LENSED_POSITION`

        Parameters
        ----------
        ps_dict : dict
            point source parameters definitions, either of `SOURCE_POSITION` or `LENSED_POSITION`

        """
        if 'ra_source' in ps_dict:
            # If the user provided ps_dict in terms of the source position, we precompute the corresponding image positions before we enter the sampling loop.
            lens_model_class = LensModel(self.kwargs_model['lens_model_list'])
            ps_class = PointSource(['SOURCE_POSITION'], lens_model_class)
            kwargs_ps_source = [ps_dict]
            ra_image, dec_image = ps_class.image_position(kwargs_ps_source, kwargs_lens)
            kwargs_image = [dict(ra_image=ra_image[0],
                                   dec_image=dec_image[0])]
            requires_reordering = True # Since the ra_image is coming out of lenstronomy, we need to reorder it to agree with TDLMC
        else:
            kwargs_image = [ps_dict]
            requires_reordering = False # If the user is providing `ra_image` inside `ps_dict`, the order is required to agree with `measured_time_delays`.
        return kwargs_image, requires_reordering

    def sample_H0(self, random_state):
        return self.H0_prior.rvs(random_state=random_state)

    def sample_kappa_ext_original(self, random_state):
        return self.kappa_ext_prior.rvs(random_state=random_state)

    def sample_kappa_ext_transformed(self, random_state):
        x = self.kappa_ext_prior.rvs(random_state=random_state)
        i = 0
        while ~np.isfinite(1.0 - 1.0/x):
            x = self.kappa_ext_prior.rvs(random_state=random_state + i)
            i += 1
        return 1.0 - 1.0/x

    def sample_aniso_param(self, random_state):
        return self.aniso_param_prior.rvs(random_state=random_state)

    def calculate_offset_from_true_image_positions(self, model_ra, model_dec, true_img_ra, true_img_dec,  increasing_dec_i, abcd_ordering_i):
        """Calculates the difference in arcsec between the (inferred or fed-in) image positions known to `H0Posterior` and the provided true image positions

        Parameters
        ----------
        true_img_ra : array-like, of length self.n_img
            ra of true image positions in TDLMC order
        true_img_dec : array-like, of length self.n_img
            dec of true image positions in TDLMC order
        
        Returns
        -------
        array-like
            offset in arcsec for each image

        """
        if self.requires_reordering:
            model_ra = h0_utils.reorder_to_tdlmc(model_ra, increasing_dec_i, abcd_ordering_i)
            model_dec = h0_utils.reorder_to_tdlmc(model_dec, increasing_dec_i, abcd_ordering_i)
        
        ra_offset = model_ra - true_img_ra
        dec_offset = model_dec - true_img_dec

        ra_offset = ra_offset.reshape(-1, 1)
        dec_offset = dec_offset.reshape(-1, 1)

        offset = np.concatenate([ra_offset, dec_offset], axis=1)
        return np.linalg.norm(offset, axis=1)

    def get_h0_sample(self, sampled_lens_model_raw, random_state):
        """Get MC samples from the H0Posterior

        Parameters
        ----------
        sampled_lens_model_raw : dict
            sampled lens model parameters, pre-formatting
        random_state : np.random.RandomState object

        Returns
        -------
        tuple of floats
            the candidate H0 and its weight

        """
        # Samples from the lens posterior are reinterpreted as samples from the lens model prior in the H0 inference stage
        lens_prior_sample = self.format_lens_model(sampled_lens_model_raw)
        kwargs_lens = lens_prior_sample['kwargs_lens']
        #lens_light_R_sersic = lens_prior_sample['lens_light_R_sersic']
        #increasing_dec_i = lens_prior_sample['increasing_dec_i']
        # Sample from respective predefined priors
        h0_candidate = self.sample_H0(random_state)
        k_ext = self.sample_kappa_ext(random_state)
        # Define cosmology
        cosmo = FlatLambdaCDM(H0=h0_candidate, Om0=self.Om0)
        # Tool for getting time delays and velocity dispersions
        td_cosmo = TDCosmography(self.z_lens, self.z_src, self.kwargs_model, cosmo_fiducial=cosmo, kwargs_lens_eqn_solver=self.kwargs_lens_eqn_solver)
        # Velocity dispersion
        # TODO: separate sampling function if vel_disp is excluded
        if self.exclude_vel_disp:
            ll_vd = 0.0
        else:
            aniso_param = self.sample_aniso_param(random_state)
            inferred_vd = self.get_velocity_dispersion(
                                                       td_cosmo, 
                                                       kwargs_lens, 
                                                       None, #FIXME: only analytic
                                                       {'aniso_param': aniso_param}, 
                                                       self.kinematics.kwargs_aperture, 
                                                       self.kinematics.kwargs_psf, 
                                                       self.kinematics.anisotropy_model, 
                                                       None,
                                                       self.kinematics.kwargs_numerics,
                                                       k_ext
                                                       )
            ll_vd = h0_utils.gaussian_ll_pdf(inferred_vd, self.measured_vd, self.measured_vd_err)
        # Time delays
        inferred_td, x_image, y_image = td_cosmo.time_delays(kwargs_lens, lens_prior_sample['kwargs_ps'], kappa_ext=k_ext)
        if len(inferred_td) > len(self.measured_td_wrt0) + 1:
            inferred_td, x_image, y_image = self.chuck_images(inferred_td, x_image, y_image)
        if lens_prior_sample['requires_reordering']:
            increasing_dec_i = np.argsort(y_image)
            inferred_td = h0_utils.reorder_to_tdlmc(inferred_td, increasing_dec_i, self.abcd_ordering_i)
        else:
            inferred_td = np.array(inferred_td)
        inferred_td_wrt0 = inferred_td[1:] - inferred_td[0]
        #print(inferred_td, self.measured_td)
        ll_td = np.sum(h0_utils.gaussian_ll_pdf(inferred_td_wrt0, self.measured_td_wrt0, self.measured_td_err))
        log_w = ll_vd + ll_td
        weight = mp.exp(log_w)
        return h0_candidate, weight

    def set_truth_lens_model(self, sampled_lens_model_raw):
        # Set once per lens
        # Samples from the lens posterior are reinterpreted as samples from the lens model prior in the H0 inference stage
        self.kwargs_model.update(dict(point_source_model_list=['SOURCE_POSITION']))
        self.lens_prior_sample = self.format_lens_model(sampled_lens_model_raw)
        cosmo = FlatLambdaCDM(H0=70.0, Om0=self.Om0) # fiducial cosmology, doesn't matter
        td_cosmo = TDCosmography(self.z_lens, self.z_src, self.kwargs_model, cosmo_fiducial=cosmo, kwargs_lens_eqn_solver=self.kwargs_lens_eqn_solver)
        _, x_image, y_image = td_cosmo.time_delays(self.lens_prior_sample['kwargs_lens'], self.lens_prior_sample['kwargs_ps'], kappa_ext=0.0)
        while len(y_image) not in [2, 4]:
            _, x_image, y_image = td_cosmo.time_delays(self.lens_prior_sample['kwargs_lens'], self.lens_prior_sample['kwargs_ps'], kappa_ext=0.0)
            #raise ValueError("Defective lens?")
        self.kwargs_model.update(dict(point_source_model_list=['LENSED_POSITION']))
        self.kwargs_image = [dict(ra_image=x_image, dec_image=y_image)]

    def get_h0_sample_truth(self, random_state):
        """Get MC samples from the H0Posterior

        Parameters
        ----------
        sampled_lens_model_raw : dict
            sampled lens model parameters, pre-formatting
        random_state : np.random.RandomState object

        Returns
        -------
        tuple of floats
            the candidate H0 and its weight

        """
        #increasing_dec_i = lens_prior_sample['increasing_dec_i']
        # Sample from respective predefined priors
        h0_candidate = self.sample_H0(random_state)
        k_ext = self.sample_kappa_ext_transformed(random_state)#self.kappa_ext #self.sample_kappa_ext(random_state) #
        # Define cosmology
        cosmo = FlatLambdaCDM(H0=h0_candidate, Om0=self.Om0)
        # Tool for getting time delays and velocity dispersions
        td_cosmo = TDCosmography(self.z_lens, self.z_src, self.kwargs_model, cosmo_fiducial=cosmo, kwargs_lens_eqn_solver=self.kwargs_lens_eqn_solver)
        # Velocity dispersion
        # TODO: separate sampling function if vel_disp is excluded
        # Time delays
        inferred_td, x_image, y_image = td_cosmo.time_delays(self.lens_prior_sample['kwargs_lens'], self.kwargs_image, kappa_ext=k_ext)
        #print(inferred_td, y_image)
        if len(inferred_td) > len(self.measured_td_wrt0) + 1:
            inferred_td, x_image, y_image = self.chuck_images(inferred_td, x_image, y_image)
            #print("after correct: ", inferred_td, y_image)
        if self.lens_prior_sample['requires_reordering']:
            increasing_dec_i = np.argsort(y_image)
            inferred_td = h0_utils.reorder_to_tdlmc(inferred_td, increasing_dec_i, self.abcd_ordering_i)
        else:
            inferred_td = np.array(inferred_td)
        inferred_td_wrt0 = inferred_td[1:] - inferred_td[0]
        #print(inferred_td_wrt0, self.measured_td_wrt0)
        ll_td = np.sum(h0_utils.gaussian_ll_pdf(inferred_td_wrt0, self.measured_td_wrt0, self.measured_td_err))
        log_w = ll_td
        weight = mp.exp(log_w)
        return h0_candidate, weight

    def chuck_images(self, inferred_td, x_image, y_image):
        """If the number of predicted images are greater than the measured, choose the images that best correspond to the measured.

        """
        # Find index or indices that must be removed.
        # Candidates 4 choose 2 (=6) or 3 choose 2 (=3)
        keep_idx = np.zeros(len(self.true_img_dec))
        for i, actual_y_i in enumerate(self.true_img_dec): # FIXME: use measured_img_dec
            keep_idx[i] = np.argmin((y_image - actual_y_i)**2.0)
        keep_idx = np.sort(keep_idx).astype(int)
        inferred_td = inferred_td[keep_idx]
        x_image = x_image[keep_idx]
        y_image = y_image[keep_idx]
        return inferred_td, x_image, y_image


"""Script to run MCMC cosmological sampling for individual lenses, using the BNN posterior

It borrows heavily from the `time-delay cosmography.ipynb` notebook in Lenstronomy Extensions, which you can find `here <https://github.com/sibirrer/lenstronomy_extensions/blob/master/lenstronomy_extensions/Notebooks/time-delay%20cosmography.ipynb>`_.

Example
-------
To run this script, pass in the path to the user-defined inference config file as the argument::
    
    $ python h0rton/infer_h0_forward_modeling.py forward_modeling.json

"""

import time
import warnings
import os
from tqdm import tqdm
import gc
from ast import literal_eval
import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from lenstronomy.Cosmo.lcdm import LCDM
from lenstronomy.Plots.model_plot import ModelPlot
import baobab.sim_utils.metadata_utils as metadata_utils
from baobab import BaobabConfig
from h0rton.configs import TrainValConfig, TestConfig
import h0rton.script_utils as script_utils
from h0rton.h0_inference import h0_utils, plotting_utils, mcmc_utils
from h0rton.h0_inference.forward_modeling_posterior import ForwardModelingPosterior
from h0rton.trainval_data import XYData

def main():
    args = script_utils.parse_inference_args()
    test_cfg = TestConfig.from_file(args.test_config_file_path)
    baobab_cfg = BaobabConfig.from_file(test_cfg.data.test_baobab_cfg_path)
    cfg = TrainValConfig.from_file(test_cfg.train_val_config_file_path)
    # Set device and default data type
    device = torch.device(test_cfg.device_type)
    if device.type == 'cuda':
        torch.set_default_tensor_type('torch.cuda.' + cfg.data.float_type)
    else:
        torch.set_default_tensor_type('torch.' + cfg.data.float_type)
    script_utils.seed_everything(test_cfg.global_seed)
    
    ############
    # Data I/O #
    ############
    # Define val data and loader
    test_data = XYData(is_train=False, 
                       Y_cols=cfg.data.Y_cols, 
                       float_type=cfg.data.float_type, 
                       define_src_pos_wrt_lens=cfg.data.define_src_pos_wrt_lens, 
                       rescale_pixels=False, 
                       rescale_pixels_type=None,
                       log_pixels=False, 
                       add_pixel_noise=cfg.data.add_pixel_noise, 
                       eff_exposure_time={"TDLMC_F160W": test_cfg.data.eff_exposure_time}, 
                       train_Y_mean=np.zeros((1, len(cfg.data.Y_cols))), 
                       train_Y_std=np.ones((1, len(cfg.data.Y_cols))), 
                       train_baobab_cfg_path=cfg.data.train_baobab_cfg_path, 
                       val_baobab_cfg_path=test_cfg.data.test_baobab_cfg_path, 
                       for_cosmology=True)
    master_truth = test_data.Y_df
    master_truth = metadata_utils.add_qphi_columns(master_truth)
    master_truth = metadata_utils.add_gamma_psi_ext_columns(master_truth)
    # Figure out how many lenses BNN will predict on (must be consecutive)
    if test_cfg.data.lens_indices is None:
        if args.lens_indices_path is None:
            # Test on all n_test lenses in the test set
            n_test = test_cfg.data.n_test 
            lens_range = range(n_test)
        else:
            # Test on the lens indices in a text file at the specified path
            lens_range = []
            with open(args.lens_indices_path, "r") as f:
                for line in f:
                    lens_range.append(int(line.strip()))
            n_test = len(lens_range)
            print("Performing H0 inference on {:d} specified lenses...".format(n_test))
    else:
        if args.lens_indices_path is None:
            # Test on the lens indices specified in the test config file
            lens_range = test_cfg.data.lens_indices
            n_test = len(lens_range)
            print("Performing H0 inference on {:d} specified lenses...".format(n_test))
        else:
            raise ValueError("Specific lens indices were specified in both the test config file and the command-line argument.")
    batch_size = max(lens_range) + 1
    test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=False, drop_last=True)
    # Output directory into which the H0 histograms and H0 samples will be saved
    out_dir = test_cfg.out_dir
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
        print("Destination folder path: {:s}".format(out_dir))
    else:
        warnings.warn("Destination folder already exists.")

    ################
    # Compile data #
    ################
    # Image data
    with torch.no_grad():
        for X_, Y_ in test_loader:
            X = X_.to(device)
            break
    X = X.detach().cpu().numpy()

    #############
    # MCMC loop #
    #############
    kwargs_lens_eqn_solver = dict(
                                  min_distance=0.05,
                                  search_window=baobab_cfg.instrument['pixel_scale']*baobab_cfg.image['num_pix'],
                                  num_iter_max=200
                                  )
    fm_posterior = ForwardModelingPosterior(kwargs_lens_eqn_solver=kwargs_lens_eqn_solver,
                                            astrometric_sigma=test_cfg.image_position_likelihood.sigma,
                                            supersampling_factor=baobab_cfg.numerics.supersampling_factor)
    # Get H0 samples for each system
    if not test_cfg.time_delay_likelihood.baobab_time_delays:
        if 'abcd_ordering_i' not in master_truth:
            raise ValueError("If the time delay measurements were not generated using Baobab, the user must specify the order of image positions in which the time delays are listed, in order of increasing dec.")

    total_progress = tqdm(total=n_test)
    realized_time_delays = pd.read_csv(test_cfg.error_model.realized_time_delays, index_col=None)
    # For each lens system...
    for i, lens_i in enumerate(lens_range):
        ###########################
        # Relevant data and prior #
        ###########################
        data_i = master_truth.iloc[lens_i].copy()
        lcdm = LCDM(z_lens=data_i['z_lens'], z_source=data_i['z_src'], flat=True)
        measured_td_wrt0 = np.array(literal_eval(realized_time_delays.iloc[lens_i]['measured_td_wrt0']))
        n_img = len(measured_td_wrt0) + 1
        #print(baobab_cfg.survey_object_dict)
        fm_posterior.set_kwargs_data_joint(
                                           image=X[lens_i, 0, :, :],
                                           measured_td=measured_td_wrt0,
                                           measured_td_sigma=test_cfg.time_delay_likelihood.sigma,
                                           survey_object_dict=baobab_cfg.survey_object_dict,
                                           eff_exposure_time=test_cfg.data.eff_exposure_time,
                                           )
        # Update solver according to number of lensed images 
        if test_cfg.numerics.solver_type == 'NONE':
            fm_posterior.kwargs_constraints['solver_type'] = 'NONE'
        else:
            fm_posterior.kwargs_constraints['solver_type'] = 'PROFILE_SHEAR' if n_img == 4 else 'ELLIPSE'
        fm_posterior.kwargs_constraints['num_point_source_list'] = [n_img]
        #print(fm_posterior.kwargs_params['point_source_model'][0][0])
        true_D_dt = lcdm.D_dt(H_0=data_i['H0'], Om0=0.3)
        # Pull truth param values and initialize walkers there
        if test_cfg.numerics.initialize_walkers_to_truth:
            fm_posterior.kwargs_lens_init = metadata_utils.get_kwargs_lens_mass(data_i)
            fm_posterior.kwargs_lens_light_init = metadata_utils.get_kwargs_lens_light(data_i)
            fm_posterior.kwargs_source_init = metadata_utils.get_kwargs_src_light(data_i)
            fm_posterior.kwargs_ps_init = metadata_utils.get_kwargs_ps_lensed(data_i)
            fm_posterior.kwargs_special_init = dict(D_dt=true_D_dt)

        ###########################
        # MCMC posterior sampling #
        ###########################
        lens_i_start_time = time.time()
        #with script_utils.HiddenPrints():
        chain_list_mcmc, kwargs_result_mcmc = fm_posterior.run_mcmc(test_cfg.numerics.mcmc)
        lens_i_end_time = time.time()
        inference_time = (lens_i_end_time - lens_i_start_time)/60.0 # min

        #############################
        # Plotting the MCMC samples #
        #############################
        # sampler_type : 'EMCEE'
        # samples_mcmc : np.array of shape `[n_mcmc_eval, n_params]`
        # param_mcmc : list of str of length n_params, the parameter names
        sampler_type, samples_mcmc, param_mcmc, _  = chain_list_mcmc[0]
        new_samples_mcmc = mcmc_utils.postprocess_mcmc_chain(kwargs_result_mcmc, 
                                                             samples_mcmc, 
                                                             fm_posterior.kwargs_model, 
                                                             fm_posterior.kwargs_params['lens_model'][2], 
                                                             fm_posterior.kwargs_params['point_source_model'][2], 
                                                             fm_posterior.kwargs_params['source_model'][2], 
                                                             fm_posterior.kwargs_params['special'][2], 
                                                             fm_posterior.kwargs_constraints,
                                                             kwargs_fixed_lens_light=fm_posterior.kwargs_params['lens_light_model'][2],
                                                             verbose=False
                                                             )
        #from lenstronomy.Plots import chain_plot
        model_plot = ModelPlot(fm_posterior.multi_band_list, 
                              fm_posterior.kwargs_model, kwargs_result_mcmc, arrow_size=0.02, cmap_string="gist_heat")
        plotting_utils.plot_forward_modeling_comparisons(model_plot, out_dir)

        # Plot D_dt histogram
        D_dt_samples = new_samples_mcmc['D_dt'].values # may contain negative values
        data_i['D_dt'] = true_D_dt
        # Export D_dt samples for this lens
        lens_inference_dict = dict(
                                   D_dt_samples=D_dt_samples, # kappa_ext=0 for these samples
                                   inference_time=inference_time,
                                   true_D_dt=true_D_dt, 
                                   )
        lens_inference_dict_save_path = os.path.join(out_dir, 'D_dt_dict_{0:04d}.npy'.format(lens_i))
        np.save(lens_inference_dict_save_path, lens_inference_dict)
        # Optionally export the MCMC samples
        if test_cfg.export.mcmc_samples:
            mcmc_samples_path = os.path.join(out_dir, 'mcmc_samples_{0:04d}.csv'.format(lens_i))
            new_samples_mcmc.to_csv(mcmc_samples_path, index=None)
        # Optionally export the D_dt histogram
        if test_cfg.export.D_dt_histogram:
            cleaned_D_dt_samples = h0_utils.remove_outliers_from_lognormal(D_dt_samples, 3)
            _ = plotting_utils.plot_D_dt_histogram(cleaned_D_dt_samples, lens_i, true_D_dt, save_dir=out_dir)
        # Optionally export the plot of MCMC chain
        if test_cfg.export.mcmc_chain:
            mcmc_chain_path = os.path.join(out_dir, 'mcmc_chain_{0:04d}.png'.format(lens_i))
            plotting_utils.plot_mcmc_chain(chain_list_mcmc, mcmc_chain_path)
        # Optionally export posterior cornerplot of select lens model parameters with D_dt
        if test_cfg.export.mcmc_corner:
            mcmc_corner_path = os.path.join(out_dir, 'mcmc_corner_{0:04d}.png'.format(lens_i))
            plotting_utils.plot_mcmc_corner(new_samples_mcmc[test_cfg.export.mcmc_cols], data_i[test_cfg.export.mcmc_cols], test_cfg.export.mcmc_col_labels, mcmc_corner_path)
        total_progress.update(1)
        gc.collect()
    realized_time_delays.to_csv(os.path.join(out_dir, 'realized_time_delays.csv'), index=None)
    total_progress.close()

if __name__ == '__main__':
    #import cProfile
    #pr = cProfile.Profile()
    #pr.enable()
    main()
    #pr.disable()
    #pr.print_stats(sort='cumtime')